#! /usr/bin/perl

#******************************************************************************
# Arianna Bisazza @ FBK-irst. April 2012
#******************************************************************************

use strict;
use open ':utf8';
binmode STDIN, ':utf8';
binmode STDOUT, ':utf8';
use Getopt::Long;


my $usage = "
-----
make-POS-mapping.pl --ttagged=TREE-TAGGED-CORPUS --dict=CORPUS-DICT --lg=en|fr OPTIONS > word-POS.map
-----
# Given a TreeTagger processed corpus and a corpus dictionary (with possible case/tokenization differences)
# creates a mapping of all the corpus tokens into their most probable POS tag
#
# Required:
# --ttagged=TREE-TAGGED-CORPUS   : NOTE that TreeTagger should be run on NON-tokenized and NON-lowercased text
# --dict=DICT: this must be the actual dictionary of the LM training corpus (may be different from tagged corpus dictionary wrt case or tokenization)
# --lg=en|fr : Supported languages are English and French
#
# Options:
# --lc       : Whether the corpus (from which the dictionary was built) is lowercased
# --lemma    : Also outputs lemma associated with most probable POS tag (in the 3rd field of the mapping)
# --tagmap=FILE : Map TreeTagger POS tags to another POS tagset (may be a partial mapping)
# --errorlog=FILE : Prints out number of tagging errors for each type 
# --debug
";

my $ttagged= '';
my $dict = '';
my $language = '';
my $lowercase = '';
my $print_lemma = '';
my $tagmap = '';
my $errorlog = '';
my $debug = '';

GetOptions ('ttagged=s' => \$ttagged,
            'dict=s'    => \$dict,
	    'lg=s'    => \$language,
	    'lc' => \$lowercase,
            'lemma' => \$print_lemma,
	    'tagmap=s' => \$tagmap,
	    'errorlog=s' => \$errorlog,
            'debug' => \$debug);

############################
# LANGUAGE DEPENDENT PARTS

my $CATCHALL_TAG="OTHER";
my $UNK_LEMMA= "<unknown>";        # The lemma of an unknown word should be itself!!
my $NUMBER_TAG= "CD";

my %EXPR_TAG = ();

if($language eq "en") {
#   $CATCHALL_TAG="NN";
    $NUMBER_TAG= "CD";
    %EXPR_TAG = (
	       "\'t","RB#not",

               "don", "VBP#do",
               "doesn", "VBZ#do",
               "didn", "VBD#do",

               "haven", "VBP#have",
               "hasn", "VBZ#have",
               "hadn", "VBD#have",

	       "aren|ain", "VBP#be",
               "isn", "VBZ#be",
               "wasn|weren", "VBD#be",

               "couldn", "MD#could",
               "wouldn", "MD#would",
               "shouldn", "MD#should",

               "\x{2019}|\x{00B4}", "POS#\'"
	       );
}
elsif ($language eq "fr") {
#    $CATCHALL_TAG="NOM";
    $NUMBER_TAG= "NUM";
    %EXPR_TAG = (
	       "d", "PRP#de",
	       "l", "DET:ART#le",
	       "n", "ADV#ne",

	       "j", "PRO:PER#je",
               "s", "PRO:PER#se",
               "m", "PRO:PER#me",

	       "qu|qu\'?", "KON#que",
               "quoiqu\'?", "KON#quoique",
               "puisqu\'?", "KON#puisque",
	       "quelqu\'?", "PRO:IND#quelque",

               "aujourd\'?|hui", "ADV#aujourd\'hui",

	       "\x{2019}|\x{00B4}", "PUN#\'"
	       );
} 
else { 
    die "Supported languages are English (en) and French (fr)\n\n$usage"; 
}


############################

#Open ttagged
$ttagged =~ s/(.*\.gz)\s*$/gzip -dc < $1|/;
open(TTAGGED, "$ttagged") or die "Cannot open \'$ttagged\': ($!) \n$usage";

#Open dict
$dict =~ s/(.*\.gz)\s*$/gzip -dc < $1|/;
open(DICT, "$dict") or die "Cannot open \'$dict\': ($!) \n$usage";
my $line = <DICT>;
if($line!~/^DICTIONARY 0/) { die "Wrong dictionary format\n\n"; }

#Open and load tagmap, if provided
my %TAGMAP;
if($tagmap ne "") {
    open(TAGMAP, "$tagmap") or die "Cannot open \'$tagmap\': ($!) \n$usage";
    while (my $line = <TAGMAP>) {
	chomp $line;
	my ($tag1,$tag2) = split(/\t+/, $line);
	if($tag1 eq "" || $tag2 eq "")  { die "Tagmap $tagmap should contain \'tag1[TAB]tag2\' pairs! Exiting\n\n"; }
	if(defined $TAGMAP{$tag1}) { die "Tagmap contains multiple entries for tag: $tag1 ! Exiting\n\\n"; }
	$TAGMAP{$tag1}=$tag2;
    }
}

# Create errorlog file, if required
if($errorlog ne "") {
    open(ERRORLOG, ">$errorlog") or die "Cannot create \'$errorlog\': ($!) \n$usage";
}

print STDERR "\n Extracting dictionary from tagged corpus...\n";

my %WORD_TAGS_LIST;
my %WORD_TAGS_LEMMA_LIST;;
my $nb_tagged_words=0;
while (my $line = <TTAGGED>) {
    $nb_tagged_words++;
    chomp $line;
    my ($word,$pos,$lemma) = split(/\t+/, $line);
    if($tagmap ne "" && defined($TAGMAP{$pos})) {$pos=$TAGMAP{$pos};}

    if($print_lemma && $lemma eq $UNK_LEMMA) {
	$lemma = $word;
    }
    if($lowercase) {
	$word = lc($word);
	$lemma = lc($lemma) if ($print_lemma);
    }
    if(!defined $WORD_TAGS_LIST{$word}) {
	my %newWord;
	$WORD_TAGS_LIST{$word}=\%newWord;
	if($print_lemma) {
	    my %newWord2;
	    $WORD_TAGS_LEMMA_LIST{$word}=\%newWord2;
	}
    }
    ${$WORD_TAGS_LIST{$word}}{$pos}++;

    if($print_lemma) {
	if(!defined ${$WORD_TAGS_LEMMA_LIST{$word}}{$pos}) {
	    my %newPos;
	    ${$WORD_TAGS_LEMMA_LIST{$word}}{$pos}=\%newPos;
	}
	${${$WORD_TAGS_LEMMA_LIST{$word}}{$pos}}{$lemma}++;
    }
}

print STDERR "Done!\n";
print STDERR "Choosing most probable tag for each word... \n";

my %WORD_TAG_MAP;
my %WORD_LEMMA_MAP;
my %TAGGING_ERRORS;
my %TAGGING_ERRORS_BY_WORD;
my $error="";
foreach my $word (sort keys (%WORD_TAGS_LIST)) {
    my @tags = sort { ${$WORD_TAGS_LIST{$word}}{$b} <=> ${$WORD_TAGS_LIST{$word}}{$a} } keys %{$WORD_TAGS_LIST{$word}};
    $WORD_TAG_MAP{$word}=$tags[0];

    if(scalar(@tags)>1) {
	foreach my $other_tag (@tags[1..$#tags]) {
	    $error=$tags[0]."=>".$other_tag;
	    $TAGGING_ERRORS{$error}+=${$WORD_TAGS_LIST{$word}}{$other_tag};
	    $TAGGING_ERRORS_BY_WORD{$word}+=${$WORD_TAGS_LIST{$word}}{$other_tag};
	}
    }

    if($print_lemma) {
	my @lemmas = sort { ${${$WORD_TAGS_LEMMA_LIST{$word}}{$tags[0]}}{$b} <=> ${${$WORD_TAGS_LEMMA_LIST{$word}}{$tags[0]}}{$a}} keys %{${$WORD_TAGS_LEMMA_LIST{$word}}{$tags[0]}};
        # assign surface form to lemma of numbers
	$lemmas[0]=$word if($lemmas[0]=~/^\@(card|ord)\@/);
	$WORD_LEMMA_MAP{$word}=$lemmas[0];
    }

    if($debug) {
	print $word, "\t", $tags[0], "\t", $WORD_LEMMA_MAP{$word}, "\n";
        foreach my $tag (@tags) { 
	    print " -> ", $word, "\t", $tag, "\t", ${$WORD_TAGS_LIST{$word}}{$tag}, "\n";
	    if($print_lemma) {
		print "    -->L--> ";
		print "MULTI! " if(scalar(keys %{${$WORD_TAGS_LEMMA_LIST{$word}}{$tag}})>1);
		foreach my $lemma (keys %{${$WORD_TAGS_LEMMA_LIST{$word}}{$tag}}) {
		    print "\t$lemma (", ${${$WORD_TAGS_LEMMA_LIST{$word}}{$tag}}{$lemma}, ")";
		}
		print "\n";
	    }
	}
	print "\n";
    }
    # check and fix some common tagging errors
    my $fixed_tag=&fix_tag($word, $tags[0]);
    if($tagmap ne "" && defined($TAGMAP{$fixed_tag})) {$fixed_tag=$TAGMAP{$fixed_tag};}
    if($fixed_tag ne $tags[0]) {
	$WORD_TAG_MAP{$word}=$fixed_tag;
        print STDERR "WARNING! Fixed tag for word -- $word -- $tags[0] => $fixed_tag \n";
    }
}

print STDERR "Done!\n\n";
print STDERR "Loading original corpus dictionary... \n";

my %DICT;
while (my $line = <DICT>) {
    chomp $line;
    my ($word,$freq) = split(/\s+/, $line);
    if($freq!~/^[0-9]+$/) { die "Dictionary $dict should contain frequencies!\n\n"; }
    $DICT{$word}=$freq;
}

print STDERR "Done!\n";
print STDERR "Checking consistency with original corpus' dictionary...\n\n";

foreach my $word (keys (%DICT)) {
    # assign some tag to words not analyzed by TreeTagger and add them to the word-tag mapping
    if(!defined $WORD_TAG_MAP{$word}) {
	my ($guessed_tag,$guessed_lemma) = split (/\#/, &guess_tag($word));
	if($tagmap ne "" && defined($TAGMAP{$guessed_tag})) {$guessed_tag=$TAGMAP{$guessed_tag};}
	$guessed_lemma="#" if($word eq "#");
        die "Lemma of $word is EMPTY!" if($guessed_lemma eq "");

	$WORD_TAG_MAP{$word} = $guessed_tag;
	$WORD_LEMMA_MAP{$word} = $guessed_lemma if($print_lemma);
	if($guessed_tag eq $CATCHALL_TAG) {
	    my $clean_word = &clean_unk_word($word);
	    if(defined $WORD_TAG_MAP{$clean_word}) { 
		$WORD_TAG_MAP{$word}=$WORD_TAG_MAP{$clean_word}; 
		$WORD_LEMMA_MAP{$word}=$WORD_LEMMA_MAP{$clean_word} if($print_lemma);
	    }
	}
	print STDERR "WARNING! Missing entry [freq: $DICT{$word}] -- $word", "\t mapped to ", $WORD_TAG_MAP{$word}, "\n";
	print STDERR "                                   with lemma: ", $WORD_LEMMA_MAP{$word}, "\n" if($print_lemma);
    }
}

print STDERR "Done!\n\n";
print STDERR "Now printing out word-tag mapping... \n\n";

print "<unk>", "\t", $CATCHALL_TAG;
print "\t", $UNK_LEMMA if($print_lemma); 
print "\n";
if($errorlog ne "") {
    print ERRORLOG "<unk>", "\t", "NAN", "\n";
}


my $word_error=0;
my $tot_word_errors=0;
foreach my $word (sort keys (%DICT)) {
    print $word, "\t", $WORD_TAG_MAP{$word};
    print "\t", $WORD_LEMMA_MAP{$word} if($print_lemma);
    print "\n";
    if($errorlog ne "") {
	$word_error=$TAGGING_ERRORS_BY_WORD{$word};
	if(!defined $word_error) { $word_error=0; }
	$tot_word_errors+=$word_error;
#	print ERRORLOG $word, "\t", $word_error, "\n";
    }
}
if($errorlog ne "") {
	print ERRORLOG "Total tagging error (by word): $tot_word_errors\n\n";
}
print STDERR "Done!\n\n";



my $tot_errors=0;
my $i=0;
print STDERR "\n********* Common tagging errors: **********\n";
foreach my $error (sort { $TAGGING_ERRORS{$b} <=> $TAGGING_ERRORS{$a} } keys %TAGGING_ERRORS) {
    $i++;
    print STDERR "err: ", $error, "\t", $TAGGING_ERRORS{$error}, "\n" if ($i<=20);
    $tot_errors+=$TAGGING_ERRORS{$error};
}
print STDERR "[...]\n\n";

my $tag_error_rate=sprintf("%.2f", $tot_errors/$nb_tagged_words *100);
print STDERR "
TAGGING ERROR STATS (ON TREE-TAGGER TYPES)
tot_errors:      $tot_errors
nb_tagged_words: $nb_tagged_words
*********************************
TAG ERROR RATE: $tag_error_rate\%
*********************************
";

#######################################

sub fix_tag {
    my $str=shift;
    my $tag=shift;
    if($language eq "en") {
        foreach my $expr (keys %EXPR_TAG) {
            if($str=~/^($expr)$/i)
            {
		$tag = $EXPR_TAG{$expr};
		$tag =~ s/\#.*//;       # strip lemma
		return $tag; 
	    }
        }
    } 
    elsif($language eq "fr") {
	##
    }
    return $tag;
}

sub guess_tag {
    my $str=shift;
    
    if($str=~/^[0-9]+$/i)
       { return "$NUMBER_TAG#$str"; }
    if($str eq "–")
       { return ":#$str"; }
    
    if($language eq "en") {
        foreach my $expr (keys %EXPR_TAG) {
            if($str=~/^($expr)$/i)
            { return $EXPR_TAG{$expr}; }
        }
    }
    elsif($language eq "fr") {
        foreach my $expr (keys %EXPR_TAG) {
            if($str=~/^($expr)$/i)
            { return $EXPR_TAG{$expr}; }
        }
    }

    return "$CATCHALL_TAG#$str";
}

sub clean_unk_word {
    my $str=shift;
    if($language eq "en") {
	##
    }
    elsif($language eq "fr") {
	# strip off final dot/quote and lookup
	$str=~s/[\.\']$//; 
	# in case hyphen, keep only first bit
	$str=~s/\-.*//;
    }

    return $str;
}
