#! /usr/bin/perl

use strict;

use open ':utf8';
binmode STDIN, ':utf8';
binmode STDOUT, ':utf8';
use Getopt::Long;


my $usage = "
-----
make-hybrid-mapping.pl --map=WORD-TAG-MAP --dict=CORPUS-DICT --minfreq=FREQ  [OPTIONS]  > word-tag.hybFREQ
-----
# Given a word-to-tag map and a dictionary with word frequencies, create a hybrid mapping
# where words occurring more than FREQ in the corpus, are mapped to themselves
#
# Required:
# --map=WORD-TAG-MAP  : must contain a tag for each entry of the dictionary DICT  (to create this use make=POS-mapping.pl)
# --dict=DICT         : original corpus' dictionary, must contain word frequencies
# --minfreq=FREQ 
#
# Options:
# --lc      : Use frequency of lowercased words as FREQ, and map frequent words to their LOWERCASED version
# --dict2=FILE  : File containing a different measure for commonality of words, to be used for minfreq threshold
# --lemma       : Query dict2 by lemma. This option requires the mapfile to contain a 3rd field with lemmas and a dict2 file with lemmas as entries
# --lemma-map   : Query dict2 by lemma (idem) AND map frequent words to their lemma, instead than to themselves
# --probWG=(1|uniform|relfreq|mostfreq)  : Outputs prob(word|class) in the 3rd field of the mapping.
#                                 Possible values: - '1' assigns 1 to all words,
#                                                  - 'uniform' assigns (1/nb_elements_of_class).
#                                                  - 'relfreq' assigns count(w)/count(g)
#                                                  - 'mostfreq' assigns count(w^g)/count(g)
# --counts-file=OUTFILE       : Prints dictionary of mapped symbols (classes) with relative counts.
# --counts-correct=[mostfreq] : Correct counts : - '1' assigns frequency of 1 to all classes,
#                                                - 'mostfreq' replaces count of each class by count of its most frequent element (if --lemma enabled, lemma frequency is used)
# --errorlog=FILE : Reads errorlog file produced by make-POS-mapping.pl and compute the tagging error of hybrid mapping
# --debug
";

my $map= '';
my $dict = '';
my $minfreq = 100;
my $lowercase = '';
my $dict2 = '';
my $use_lemma = '';
my $map_to_lemma = '';
my $probWG = '';
my $counts_file = '';
my $counts_correct = '';
my $errorlog = '';
my $debug = '';

GetOptions ('map=s' => \$map,
	    'dict=s'    => \$dict,
            'minfreq=f' => \$minfreq,
            'lc' => \$lowercase,
            'dict2=s'  => \$dict2,
	    'lemma'    => \$use_lemma, 
            'lemma-map'    => \$map_to_lemma,
            'probWG=s' => \$probWG,
            'counts-file=s' => \$counts_file,
            'counts-correct=s' => \$counts_correct,
            'errorlog=s' => \$errorlog,
            'debug' => \$debug);

my $UNKNOWN = "<unk>";
my $UNKNOWN_TAG = "<unk>";
my $UNKNOWN_PROB = "0.000001";
#my $CATCHALL_TAG="NN";
#my $CATCHALL_TAG="OTHER";

# Check option compatibility
if($map_to_lemma) {
    $use_lemma=1;
    print STDERR "\nMapping frequent words to their LEMMA !!\n";
}
if($use_lemma && $dict2 eq "") { die "--lemma option requires a dict2 file with lemmas as entries! Exiting.\n"; }
if($use_lemma) {
    print STDERR "\nUsing LEMMA statistics for threshold...\n"; 
}

# Check value of 'probWG' option
($probWG=~/^(|1|uniform|relfreq|mostfreq)$/) or die "Undefined value for option \'probWG\': $probWG \n$usage";
if($probWG ne "") {
    print STDERR "\nPrinting out prob(word|class). Mode: $probWG\n";
}

# Check value of 'counts-correct' option
my $FH_COUNTS;
($counts_correct=~/^(|1|mostfreq)$/) or die "Undefined value for option \'counts-correct\': $counts_correct \n$usage";
if($counts_file ne "") {
    print STDERR "\nPrinting class counts to file $counts_file\n";
    if($counts_correct ne "") {
	print STDERR "\nCount correction mode: $counts_correct\n";
    }
    open($FH_COUNTS, ">$counts_file") or die "Cannot open $counts_file for writing: ($!) \n$usage";
}

#Open map
$map =~ s/(.*\.gz)\s*$/gzip -dc < $1|/;
open(MAP, "$map") or die "Cannot open \'$map\': ($!) \n$usage";

#Open dict
$dict =~ s/(.*\.gz)\s*$/gzip -dc < $1|/;
open(DICT, "$dict") or die "Cannot open \'$dict\': ($!) \n$usage";
my $line = <DICT>;
if($line!~/^DICTIONARY 0/) { die "Wrong dictionary format\n\n"; }

#Open dict2
if($dict2 ne "") {
    $dict2 =~ s/(.*\.gz)\s*$/gzip -dc < $1|/;
    open(DICT2, "$dict2") or die "Cannot open \'$dict2\': ($!) \n$usage";
    print STDERR "Using values $dict2 for threshold!!\n\n";
}

# Open errorlog file, if provided
if($errorlog ne "") {
    open(ERRORLOG, "$errorlog") or die "Cannot open \'$errorlog\': ($!) \n$usage";
}

##############################################

print STDERR "\nHYBRID MODE: words with frequency>=$minfreq are mapped to themselves\n";
if($lowercase) { 
    print STDERR "
      ####################################
        ==> CASE-INSENSITIVE MAPPING <==
      ####################################
"; }

################################################
my %DICT2;
#my %DICT2_lc;
if($dict2 ne "") {
    while (my $line = <DICT2>) {
	chomp $line;
	my ($word,$freq) = split(/\s+/, $line);
	if($freq!~/^[0-9\.\-]+$/) { die "Dictionary $dict should contain real numbers! Exiting\n\n"; }
	if($lowercase) { 
	    $word = lc($word);
	}
	$DICT2{$word}+=$freq;
    }
}

print STDERR "\n Loading word-tag map...\n";

my %WORD_TAG_MAP;
my %WORD_LEMMA_MAP;
my %WORD_ERROR_COUNT;
while (my $line = <MAP>) {
    chomp $line;
    my ($word,$tag,$lemma) = split(/\t+/, $line);
    if($word eq "" || $tag eq "")  { die "Map $map should contain \'word[TAB]tag\' pairs! Exiting\n\n"; }

    if($use_lemma) {
	if($lemma eq "") { die "Map $map should contain \'word[TAB]tag[TAB]lemma\' triplets, with --lemma enabled! Exiting\n\n"; }
	$WORD_LEMMA_MAP{$word}=$lemma;
    }
    if($word eq $UNKNOWN) {
	$UNKNOWN_TAG=$tag;
    }

    $WORD_TAG_MAP{$word}=$tag;
    if($errorlog ne "") {
	my $eline = <ERRORLOG>;
	chomp $eline;
	my ($word2,$nb_errors) = split(/\t+/, $eline);
	if($word2 ne $word || $nb_errors eq "")  { die "Error log $errorlog should contain \'word[TAB]nb_errors\' pairs and same entries as MAPFILE! \n$eline \nExiting\n\n"; }
	$WORD_ERROR_COUNT{$word}=$nb_errors;
    }
}

print STDERR "Done!\n";
print STDERR "Loading original corpus dictionary... \n";

my %DICT;
my %DICT_lc;
while (my $line = <DICT>) {
    chomp $line;
    my ($word,$freq) = split(/\s+/, $line);
    if($freq!~/^[0-9]+$/) { die "Dictionary $dict should contain frequencies! Exiting\n\n"; }
    $DICT{$word}=$freq;
    if($lowercase)
    { $DICT_lc{lc($word)}+=$freq; }
}
#foreach my $word (sort keys (%DICT_lc))
#{ print "DEBUGlc $word\t$DICT_lc{$word}\n"; }


print STDERR "Done!\n\n";
print STDERR "Making hybrid mapping...\n\n";
my $rh_QUERIED_DICT = \%DICT;
if($dict2) 
  { $rh_QUERIED_DICT = \%DICT2; }
elsif($lowercase)
  { $rh_QUERIED_DICT = \%DICT_lc; }

my %HYB_WORD_TAG_MAP;
my %CCOUNTS_LEMMA;
my $nbFreqWords=0;  my $nbWords=0;
my $nbFreqTokens=0; my $nbTokens=0;
my $processed=0;
my $tagging_errors;
my $freq; my $freq_for_thresh; 
my $mapped_word=""; my $queried_word="";
foreach my $word (keys (%DICT)) {
    $processed=0;
    $freq=$DICT{$word};
    $nbWords++; $nbTokens+=$freq;
    $queried_word=$mapped_word=$word;
    if($use_lemma) {
	$queried_word=$WORD_LEMMA_MAP{$word};
    }
    if($map_to_lemma) {
	$mapped_word=$WORD_LEMMA_MAP{$word};
    }
    if($lowercase){
	$mapped_word=lc($mapped_word);
	$queried_word=lc($queried_word);
    }

    $freq_for_thresh=${$rh_QUERIED_DICT}{$queried_word};
    if(!defined $freq_for_thresh) { die "This entry ($queried_word) should be in the dictionary used for frequency! \nExiting\n"; }    

    if($freq_for_thresh>=$minfreq) {
	$nbFreqWords++; 
	$nbFreqTokens+=$freq;
	$HYB_WORD_TAG_MAP{$word}=$mapped_word;
	$processed=1;
	if($debug) {
	    print STDERR "above thresh: $freq_for_thresh --\t $word => $mapped_word \n";
	}
    }
    next if($processed);
    if(!defined $WORD_TAG_MAP{$word}) { die "This word should be in the map! $word \nExiting\n"; }
    $HYB_WORD_TAG_MAP{$word}=$WORD_TAG_MAP{$word};
    if($errorlog ne "") {
	#print STDERR "";
	$tagging_errors+=$WORD_ERROR_COUNT{$word};
    }
}

##### Assign uniform probabilities to each class : prob(word|class)=1/nb_elements_of_class
my %CLASS_PROB;
if($probWG eq "uniform") {
    foreach my $word (keys (%HYB_WORD_TAG_MAP)) {
	$CLASS_PROB{$HYB_WORD_TAG_MAP{$word}}++;
    }
    foreach my $tag (keys (%CLASS_PROB)) {
	$CLASS_PROB{$tag} = sprintf("%.6f", 1/$CLASS_PROB{$tag});
    }
}

##### Assign max-likelihood (rel.frequency-based) probabilities to each word : prob(word|class)=count(word)/count(class)
my %CLASS_COUNT;
my %WORD_PROB;   # note: a simple hash is enough, since by definition each word occurs only in one class
my %CLASS_HIGHEST_FREQ;
if($probWG =~ /(relfreq|mostfreq)/) {
    foreach my $word (keys (%HYB_WORD_TAG_MAP)) {
	my $class=$HYB_WORD_TAG_MAP{$word};
        $CLASS_COUNT{$class}+=$DICT{$word};
	if($DICT{$word}>$CLASS_HIGHEST_FREQ{$class}) {
	    $CLASS_HIGHEST_FREQ{$class}=$DICT{$word};
	}
    }
    foreach my $word (keys (%HYB_WORD_TAG_MAP)) {
	my $class=$HYB_WORD_TAG_MAP{$word};
	#if(defined $CLASS_COUNT{$WORD_TAG_MAP{$word}} || $CLASS_COUNT{$WORD_TAG_MAP{$word}}!=0)
	if($probWG eq "relfreq") {
	    $WORD_PROB{$word} = sprintf("%.6f", $DICT{$word}/$CLASS_COUNT{$class});
	} elsif($probWG eq "mostfreq") {
	    $WORD_PROB{$word} = sprintf("%.6f", $CLASS_HIGHEST_FREQ{$class}/$CLASS_COUNT{$class});
	}
    }
}

print STDERR "Done!\n\n";
print STDERR "Now printing out hybrid word-tag mapping... \n\n";

print $UNKNOWN, "\t", $UNKNOWN_TAG;
if($probWG ne "") {
    print "\t", $UNKNOWN_PROB;
}
print "\n";

foreach my $word (sort keys (%HYB_WORD_TAG_MAP)) {
    print $word, "\t", $HYB_WORD_TAG_MAP{$word};
    if($probWG ne "" && $HYB_WORD_TAG_MAP{$word} eq $UNKNOWN_TAG) {
	print "\t", $UNKNOWN_PROB, "\n";
	next;
    }
    if($probWG eq "1") { print "\t", 1; }
    elsif($probWG eq "uniform") { print "\t", $CLASS_PROB{$HYB_WORD_TAG_MAP{$word}}; }
    elsif($probWG =~ /(relfreq|mostfreq)/) { print "\t", $WORD_PROB{$word}; }
    print "\n";
}

print STDERR "Done!\n\n";
print STDERR "Now printing out counts of classes resulting from mapping... \n\n";

my %CLASS_COUNT;
my %CLASS_MFREQ;
##### Compute and print out counts of classes resulting from mapping
if($counts_file ne "") {
    foreach my $word (keys (%HYB_WORD_TAG_MAP)) {
	if($counts_correct eq "") {
	    if(defined($DICT{$word})) {
		$CLASS_COUNT{$HYB_WORD_TAG_MAP{$word}}+=$DICT{$word};
	    }
	} elsif($counts_correct eq "1") {
	    $CLASS_COUNT{$HYB_WORD_TAG_MAP{$word}}=1;
	} elsif($counts_correct eq "mostfreq") {
	    # assign this word if more frequent than current word
	    if($DICT{$word}>$CLASS_COUNT{$HYB_WORD_TAG_MAP{$word}}) { 
		$CLASS_COUNT{$HYB_WORD_TAG_MAP{$word}}=$DICT{$word};
		$CLASS_MFREQ{$HYB_WORD_TAG_MAP{$word}}=$word;
	    }
	}
    }
    foreach my $class (sort keys (%CLASS_COUNT)) {
	if($counts_correct eq "mostfreq" && $use_lemma) {
	    my $queried_word=$CLASS_MFREQ{$class};
	    $queried_word=$WORD_LEMMA_MAP{$queried_word} if($use_lemma);
	    $queried_word=lc($queried_word) if($lowercase);
	    if($debug) {
		print $FH_COUNTS "MFREQ($class)=", $CLASS_MFREQ{$class}, " Querying dict with token: $queried_word\n"; 
	    }
	    print $FH_COUNTS $class, "\t", ${$rh_QUERIED_DICT}{$queried_word}, "\n"; 
	}else{
	    print $FH_COUNTS $class, "\t", $CLASS_COUNT{$class}, "\n";
	}
    }
}


print STDERR "Done!\n";

my $freqWords_percent= sprintf("%.2f", $nbFreqWords/$nbWords*100);
my $freqTokens_percent=sprintf("%.2f", $nbFreqTokens/$nbTokens*100);
   
my $freqtype="";
if($lowercase) { $freqtype="lowercased"; }

print STDERR "----------------------------------------------------------------------------------\n";
print STDERR "$freqWords_percent% of entries have $freqtype frequency>=$minfreq ";
print STDERR "covering $freqTokens_percent% of corpus tokens\n";
print STDERR "----------------------------------------------------------------------------------\n\n";

if($errorlog ne "") {
    print STDERR "\n*************************************************************\n";
    print STDERR "  [Hybrid mapping] Total tagging errors on POS-mapped words: ", $tagging_errors;
    print STDERR "\n*************************************************************\n";
}

